/*$Id: acre_set_context.c,v 1.1 2004/04/30 Joel M. Malard Exp d3j540 $ */

#include "tao_solver.h"
#include "petscblaslapack.h"
#include "petscmat.h"
#include "petscksp.h"
#include "acre.h"
#include <assert.h>
#include <mpi.h>
#include "copyright.h"
#undef __FUNC__
#define __FUNC__ "acre_set_context"
int acre_set_context(acre_cntxt *cntxt, Vec iVec)
{
	/* Purpose:
	 * returns immediately if the input vector is the same
	 * as before; otherwise make a serial copy of the input
	 * vector, fill out the cholesky factor and its
	 * transpose (the latter matrix is needed when
	 * KSPSolveTranspose did not work with all PETSc solvers),
	 * reset the linear system solver contexts for L and its
	 * transpose, if L is not singular:
	 * 	precompute B such as L*B=A,
	 * 	precompute the skinny QR factorization of B,
	 *	precompute w in Lw=b,
	 * 	precompute v = L^Tu in Ax+Cu=b,
	 * 	precompute u in Ax+Bu=b,
	 * 	save the new input vector.
	 * When multiple groups of processes exist, iVec might be
	 * obtained from process 0.
	 * 
	 * Return Value: 0 if this is a new vector and all went
	 * well, 1 if the input vector is the same as last time
	 * SetContext was called, -1 if an error was detected.
	 */
	int ierr;
	int nq;
	PetscScalar mone = -1.0;
	Vec z;
	Vec y;
	PetscTruth sameVec;
	static int ncall=0;
	int here, cank, rank, rize;

	/* cntxt->pVec holds a local copy of the variance
	 * components, so nq must not be too large
	 */
	if ( ncall>0 )
	{
		ierr = VecEqual(iVec, cntxt->pVec, &sameVec);	CHKERRQ(ierr);
	}
	if ( ncall>0 && sameVec==PETSC_TRUE) return 1;

	HELLO("acre_set_context start");
	/* This is a new input vector: invalidate the old gradient and
	 * likelihood.
	 */
	cntxt->bgSet = 0;
	cntxt->bfSet = 0;

	/* Save the current vector of components */
	ierr = VecCopy( iVec, cntxt->pVec);			CHKERRQ(ierr);

	/* find which process this is. */
	ierr = MPI_Comm_rank(MPI_COMM_WORLD,&here);
	ierr = MPI_Comm_rank(PETSC_COMM_WORLD,&cank);
	ierr = MPI_Comm_rank(cntxt->rcomm,&rank);
	ierr = MPI_Comm_size(cntxt->rcomm,&rize);

	/* Redefine the matrix L and its transpose */
	ierr = acre_set_cholesky(cntxt, iVec);			CHKERRQ(ierr);

	/* Reset the linear system solver context */
	ierr = KSPSetOperators(cntxt->sles, cntxt->L,
		cntxt->L, SAME_NONZERO_PATTERN );		CHKERRQ(ierr);

	/* @@@ */
	ierr = KSPSetTolerances( cntxt->sles, ACRE_EPSILON,
		ACRE_EPSILON, PETSC_DEFAULT,
		PETSC_DEFAULT);					CHKERRQ(ierr);
	/* @@@ */

	ierr = KSPSetFromOptions(cntxt->sles);			CHKERRQ(ierr);

	ierr = KSPSetOperators(cntxt->stes, cntxt->LTr,
		cntxt->LTr, SAME_NONZERO_PATTERN );		CHKERRQ(ierr);

	/* @@@ */
	ierr = KSPSetTolerances( cntxt->stes, ACRE_EPSILON,
		ACRE_EPSILON, PETSC_DEFAULT,
		PETSC_DEFAULT);					CHKERRQ(ierr);
	/* @@@ */

	ierr = KSPSetFromOptions(cntxt->stes);			CHKERRQ(ierr);

	/* Get a local copy of the input vector */
	ierr = VecGetSize(iVec, &nq);				CHKERRQ(ierr);

	/* Precompute B such that LB=A, assume L is 
	 * not singular.
	 */
	acre_sles_mat( cntxt, cntxt->A, cntxt->B,
		cntxt->v, ACRE_NO_TRANSPOSE);

	/* Precompute the skinny QR factorization of B */
	ierr = MatCopy( cntxt->B, cntxt->Q,
		SAME_NONZERO_PATTERN);				CHKERRQ(ierr);
	acre_mgs( cntxt, cntxt->Q, cntxt->R);

	/* Compute w in  Lw= b */
	ierr = KSPSetRhs(cntxt->sles, cntxt->b);		CHKERRQ(ierr);
	ierr = KSPSetSolution(cntxt->sles, cntxt->w);		CHKERRQ(ierr);
	ierr = KSPSolve(cntxt->sles);				CHKERRQ(ierr);

	/* Precompute cntxt->v=L^Tu of Ax+Cu=b */
	ierr = VecCreateMPI(PETSC_COMM_WORLD,PETSC_DECIDE,
		cntxt->rn,&z);	CHKERRQ(ierr);
	ierr = MatMultTranspose(cntxt->Q,cntxt->w,z);		CHKERRQ(ierr);
  	ierr = VecDuplicate( cntxt->b,  &y);			CHKERRQ(ierr);
	ierr = MatMult(cntxt->Q,z,y);				CHKERRQ(ierr);
	ierr = acre_vector_combine(cntxt, y);			CHKERRQ(ierr);
	ierr = VecCopy(cntxt->w,cntxt->v);			CHKERRQ(ierr);
	ierr = VecAXPY(&mone,y,cntxt->v);			CHKERRQ(ierr);

	ierr = KSPSetRhs( cntxt->stes, cntxt->v);		CHKERRQ(ierr);
	ierr = KSPSetSolution( cntxt->stes, cntxt->u);		CHKERRQ(ierr);
	ierr = KSPSolve ( cntxt->stes );			CHKERRQ(ierr);

	ierr = VecDestroy(z);					CHKERRQ(ierr);
	ierr = VecDestroy(y);					CHKERRQ(ierr);

	HELLO("acre_set_context exit");

	return 0;

}

