/* $Id: acre_start.c,v 1.1 2004/04/30 05:16:51 Joel M. Malard Exp d3j540 $ */

#include "tao_solver.h"
#include "petscblaslapack.h"
#include "petsclog.h"
#include "acre.h"
#include "petsclog.h"
#include <assert.h>
#include <mpi.h>
#include "copyright.h"
#undef __FUNC__
#define __FUNC__ "acre_start"
/* Sets up the application context */
int acre_start(Mat A, Vec b, Mat sL, Vec c, int qc, acre_cntxt *cntxt)
{
	/* Purpose:
	i.   Allocate storage for the internal data
	ii.  Set up the linear solver context
	iii. Set VecScatter to allow local copies of the component vector
	iv.  Set data structure to speed-up the computation of the 
		determinant of L.

	Note: acr->ofcn must be set by the user before calling acre_start.

	Arguments:
	If ACRE_DESIGN is defined then the structural matrix sL must be
	defined before ACRE_Start is called.

	*/
	int ierr, here, cank, rank, rize;
	int m, n, nq;
	PetscScalar zero=0.0;
	int i;

	HELLO("acre_start enter");

	/* first define a communicator for block rows */
	cntxt->qc = qc;

	ierr = MPI_Comm_rank(MPI_COMM_WORLD,&here);
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_rank(PETSC_COMM_WORLD,&cank);
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_split(MPI_COMM_WORLD, cank, here,
		&cntxt->rcomm);
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_rank(cntxt->rcomm, &rank);
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_size(cntxt->rcomm, &rize);
		assert(ierr==MPI_SUCCESS); 

	/* The matrix A defines the 2D data distribution */
	cntxt->A = A;
	ierr = MatGetSize(A,&cntxt->m,&cntxt->rn);		CHKERRQ(ierr);
	ierr = MPI_Allreduce( &cntxt->rn, &cntxt->n, 
		1, MPI_INT, MPI_SUM, cntxt->rcomm);
		assert(ierr==MPI_SUCCESS);

	/* record the starting column of A for each 
	 * process column.
	 */
	ierr = PetscMalloc((rize+1)*sizeof(int),
			&cntxt->rstart );			CHKERRQ(ierr);

	for ( i=0; i<qc; i++ )
	{
		int h = i*(cntxt->n/qc);
		int k = cntxt->n%qc;
		cntxt->rstart[i] = h + ((i<k) ? i : k);
	}
	cntxt->rstart[qc] = cntxt->n;

	/* now register the structural matrix */
	cntxt->sL = sL;
	ierr = MatGetSize(sL,&m,&cntxt->nq);			CHKERRQ(ierr);
	if(m!=cntxt->m)CHKERRQ(-1);

	/*** Allocate storage for  the CHolesky factors of
	 *** the covariance matrix and its transpose. 
	 *** Both factors are allocated because the Petsc
	 *** 1.1.0 does not support systematically 
	 *** linear system solution with an implicit
	 *** transpose matrix. One way around might be 
	 *** with shell matrices.
	 ***/

	ierr = MatCreateMPIAIJ( PETSC_COMM_WORLD,
		PETSC_DECIDE, PETSC_DECIDE, cntxt->m,
		cntxt->m, cntxt->nq, (int *) 0,
		cntxt->nq, (int *) 0, &cntxt->L);		CHKERRQ(ierr);

#if !defined(USE_PETSC_TRANSPOSE)
	ierr = MatCreateMPIAIJ( PETSC_COMM_WORLD,
		PETSC_DECIDE, PETSC_DECIDE, cntxt->m,
		cntxt->m, cntxt->nq, (int *) 0,
		cntxt->nq, (int *) 0, &cntxt->LTr);		CHKERRQ(ierr);
#endif

	ierr = MatGetSize(cntxt->L,&m,&n);			CHKERRQ(ierr);
	if(m!=cntxt->m || n!=cntxt->m)CHKERRQ(-1);

	/* register the problem vector */
	cntxt->b = b;
	ierr = VecGetSize(b,&m);				CHKERRQ(ierr);
	if(m!=cntxt->m)CHKERRQ(-1);

	/* register the solution vector */
	ierr = VecGetSize(c,&nq);				CHKERRQ(ierr);
	if(nq!=cntxt->nq)CHKERRQ(-1);

	/* create solver contexts for L and its transpose */
	ierr = KSPCreate (PETSC_COMM_WORLD, &cntxt->sles);	CHKERRQ(ierr);
	ierr = KSPSetTolerances( cntxt->sles, ACRE_EPSILON,
		ACRE_EPSILON, PETSC_DEFAULT, PETSC_DEFAULT);	CHKERRQ(ierr);
	ierr = KSPSetFromOptions(cntxt->sles);			CHKERRQ(ierr);

#if !defined(USE_PETSC_TRANSPOSE)
	ierr = KSPCreate (PETSC_COMM_WORLD, &cntxt->stes);	CHKERRQ(ierr);
	ierr = KSPSetTolerances( cntxt->stes, ACRE_EPSILON,
		ACRE_EPSILON, PETSC_DEFAULT, PETSC_DEFAULT);	CHKERRQ(ierr);
	ierr = KSPSetFromOptions(cntxt->stes);			CHKERRQ(ierr);
#endif

	/* precompute locations of diagonal elements in L */
	ierr = VecCreateSeq(PETSC_COMM_SELF, nq, 
		&cntxt->sLDiag); 				CHKERRQ(ierr);
	ierr = acre_detl(cntxt->sL, cntxt->sLDiag);		CHKERRQ(ierr);

	/* keep track of last input vector in pVec */
	ierr = VecDuplicate( c, &cntxt->pVec );			CHKERRQ(ierr);
	ierr = VecSet( &zero, cntxt->pVec );			CHKERRQ(ierr);

	/* keep track of corresponding gradient in gVec */
	ierr = VecDuplicate( c, &cntxt->gVec );			CHKERRQ(ierr);
	ierr = VecSet( &zero, cntxt->gVec );			CHKERRQ(ierr);
	cntxt->bgSet = 0;

	/* ditto for the likelihood */
	cntxt->fScal = 0;
	cntxt->bfSet = 0;

	/* ditto for the vectors u and v=L^T*u */
	ierr = VecDuplicate( b, &cntxt->w);			CHKERRQ(ierr);
	ierr = VecDuplicate( b, &cntxt->v);			CHKERRQ(ierr);
	ierr = VecSet( &zero, cntxt->v);			CHKERRQ(ierr);
	ierr = VecDuplicate( b, &cntxt->u);			CHKERRQ(ierr);

	/* reserve storage for the solution B of L*B=A */
	ierr = MatCreateMPIDense(
		PETSC_COMM_WORLD, PETSC_DECIDE,
		PETSC_DECIDE, cntxt->m, cntxt->rn,
		TAO_NULL, &cntxt->B);				CHKERRQ(ierr);

	ierr = MatZeroEntries( cntxt->B);			CHKERRQ(ierr);
	ierr = MatAssemblyBegin( cntxt->B, 
		MAT_FINAL_ASSEMBLY );				CHKERRQ(ierr);
	ierr = MatAssemblyEnd( cntxt->B, MAT_FINAL_ASSEMBLY );	CHKERRQ(ierr);

	/* allocate storage for the QR factorization of B */
	ierr = MatCreateMPIDense(
		PETSC_COMM_WORLD, PETSC_DECIDE,
		PETSC_DECIDE, cntxt->m, cntxt->rn,
		TAO_NULL, &cntxt->Q);				CHKERRQ(ierr);
	ierr = MatZeroEntries( cntxt->Q);			CHKERRQ(ierr);
	ierr = MatAssemblyBegin( cntxt->Q, 
		MAT_FINAL_ASSEMBLY );				CHKERRQ(ierr);
	ierr = MatAssemblyEnd( cntxt->Q, MAT_FINAL_ASSEMBLY );	CHKERRQ(ierr);

	ierr = MatCreateSeqDense( PETSC_COMM_SELF,
		cntxt->n,cntxt->n,PETSC_NULL,&(cntxt->R));      CHKERRQ(ierr);
	ierr = MatZeroEntries( cntxt->R);			CHKERRQ(ierr);
	ierr = MatAssemblyBegin( cntxt->R, 
		MAT_FINAL_ASSEMBLY );				CHKERRQ(ierr);
	ierr = MatAssemblyEnd( cntxt->R, MAT_FINAL_ASSEMBLY );	CHKERRQ(ierr);

	cntxt->singf = 0;

	HELLO("acre_start exit");

	return 0;

}

